/*
 * Copyright (C) 2001 Adrian Schroeter <adrian@suse.de>
 */

#include <kfiledialog.h>
#include <kaboutdata.h>
#include <kapplication.h>
#include <kcmdlineargs.h>
#include <kglobalsettings.h>
#include <klocale.h>
#include <qstringlist.h>
#include <qsignal.h>
#include <stdlib.h>

#include <tools.h>

static const char *description =
    I18N_NOOP("Front-end for the KDE file dialog");

static const char *version = "0.1";

static KCmdLineOptions options[] =
{
    { "file", I18N_NOOP( "Set dialog mode to single file (default)" ), 0 },
    { "files", I18N_NOOP( "Set dialog mode to multiple file" ), 0 },
    { "directory", I18N_NOOP( "Set dialog mode to directory" ), 0 },
    { "selected", I18N_NOOP( "Print selected files" ), 0 },
    { "remote-files", I18N_NOOP( "Accept nonlocal files" ), 0 },
    { "+[DIRECTORY]", I18N_NOOP( "Set initial directory" ), 0 },
    { "+[FILTER}", I18N_NOOP( "Set filter" ), 0 },
    { 0, 0, 0 }
};

Tools::Tools( bool _allowRemote ){
    oldFile = QString::null;
    allowRemote = _allowRemote;
}

void Tools::selectedFile(const QString &file){
    if ( file != oldFile ){
       if ( allowRemote )
	  printf("%s\n", file.ascii());
	else
	  printf("%s\n", file.mid(5).ascii());
       oldFile=file;
    }
};

void Tools::reg(QWidget *w){
      connect( w, SIGNAL(fileHighlighted(const QString&)), this, SLOT(selectedFile(const QString&)));
};

int main(int argc, char **argv)
{
    KAboutData about("kfiledialog", I18N_NOOP("kfiledialog"), version, description,
                     KAboutData::License_GPL, "(C) 2001 Adrian Schroeter", 0, 0, "adrian@suse.de");
    about.addAuthor( "Adrian Schroeter", 0, "adrian@suse.de" );

    QString file = QString::null;
    QStringList files;

    QString startDir;
    QString filter("*");

    KCmdLineArgs::init(argc, argv, &about);
    KCmdLineArgs::addCmdLineOptions(options);

    KCmdLineArgs *args = KCmdLineArgs::parsedArgs();
    KApplication *app  = new KApplication();
    Tools *tools       = new Tools(args->isSet("remote-files"));
    KFileDialog *kfd;

    if ( args->count() > 0 )
       startDir = QFile::decodeName( args->arg(0) );
    else
       startDir = KGlobalSettings::documentPath();
    if ( args->count() > 1 )
       filter   = QFile::decodeName( args->arg(1) );

    kfd = new KFileDialog( startDir, filter, 0, "kfiledialog", true );

    if ( args->isSet("selected") )
       tools->reg(kfd);

    kfd->setOperationMode( KFileDialog::Opening );
    if ( args->isSet("files") )
	kfd->setMode(KFile::Files | KFile::LocalOnly);
    else if ( args->isSet("directory") )
	kfd->setMode(KFile::Files | KFile::LocalOnly);
    else
	kfd->setMode(KFile::File | KFile::LocalOnly);

//   kfd->setCaption(caption.isNull() ? i18n("Open") : caption);
//    kfd->ops->clearHistory();
    kfd->exec();

    if ( args->isSet("files") )
        files = kfd->selectedFiles();
    else
        file  = kfd->selectedFile();

    args->clear();
    if ( files.count() ){
	QStringList::iterator it;
	for ( it = files.begin(); it != files.end(); ++it )
		printf("%s\n", (*it).ascii() );
    	return 0;
    }

    if ( !file.isEmpty() ){
	printf("%s\n", file.latin1());
    	return 0;
    }

    return -1;
}

#include <tools.moc>

